;;; sccs-complete.el --- Completion of "SCCS/s.*" files

;; Copyright (C) 2003 Bryan O'Sullivan

;; Author: Bryan O'Sullivan <bos@serpentine.com>

;; $Id: sccs-complete.el 1.3 03/05/15 16:42:37-07:00 bos@serpentine.internal.keyresearch.com $

;; sccs-complete.el ("this file") is free software; you can
;; redistribute it and/or modify it under the terms of the GNU General
;; Public License as published by the Free Software Foundation; either
;; version 2, or (at your option) any later version.

;; This file is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this file, GNU Emacs, or XEmacs; see the file COPYING
;; (`C-h C-l').  If not, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;; Commentary:

;; Although Emacs knows how to do a "get" of a file "foo" if it
;; doesn't exist, but an "SCCS/s.foo" file exists, it does not perform
;; file name completion on those "SCCS/s.*" files.  This is pretty
;; irritating if you deal with tools that use SCCS for storage, such
;; as BitKeeper.

;; This package advises the Emacs file name completion functions, so
;; that they complete on "s.*" files in an SCCS subdirectory, as well
;; as the usual files.

;; At the moment, this package only works under XEmacs.  It loads
;; under GNU Emacs, but GNU Emacs doesn't seem to call back into Lisp
;; when completing file names for the likes of find-file.

(require 'advice)
(require 'cl)

(defun sccs-make-completions (partial-filename directory &optional existing)
  "Return all completions of `s.PARTIAL-FILENAME' in `DIRECTORY/SCCS'.
This is a list of all file names in `DIRECTORY/SCCS' which begin with
`s.PARTIAL-FILENAME', with the `SCCS/s.' removed from each name."
  (when (or (not directory) (eq (length directory) 0))
    (setq directory "."))
  (let ((sccs-subdir (concat directory "/SCCS"))
	(sccs-completing t)
	names)
    (when (and sccs-completing (file-directory-p sccs-subdir))
      (setq sccs-completing nil)
      (dolist (sfile (mapcar (function (lambda (x) (substring x 2)))
			     (file-name-all-completions 
			      (concat "s." partial-filename)
			      sccs-subdir)))
	(unless (member sfile existing)
	  (setq names (cons sfile names))))
      names)))

(defadvice file-name-all-completions (after sccs-complete-files activate)
  "Complete on `s.' files in the SCCS subdirectory, as well as entries
in the current directory."
  (let (sccs-completions)
    (let ((sc-file (ad-get-arg 0))
	  (sc-dir (ad-get-arg 1)))
      (dolist (item (sccs-make-completions sc-file sc-dir
					   ad-return-value))
	(unless (member item ad-return-value)
	  (setq sccs-completions (cons item sccs-completions)))))
    (setq ad-return-value (append sccs-completions ad-return-value))))

(defun sccs-ignore-completions (completions)
  "Filter the list of COMPLETIONS, ignoring appropriate items.
The items to ignore are strings that end with any of the suffixes in
`completion-ignored-extensions'."
  (let (comps)
    (dolist (item completions)
      (let ((ignore completion-ignored-extensions)
	    ignore-item)
	(while (and ignore (not ignore-item))	   
	  (let* ((ign (car ignore))
		 (ign-len (length ign)))
	    (when (and (> (length item) ign-len)
		       (string-equal ign (substring item (- ign-len))))
	      (setq ignore-item t))
	    (setq ignore (cdr ignore))))
	(unless ignore-item
	  (setq comps (cons item comps)))))
    comps))
  
(defadvice file-name-completion (after sccs-complete-files activate)
  "Complete on `s.' files in the SCCS subdirectory, as well as entries
in the current directory."
  (let* ((sc-file (ad-get-arg 0))
	 (sc-dir (ad-get-arg 1))
	 (rv-len (length ad-return-value))
	 (pf-len (length sc-file)))
    (dolist (item (sccs-ignore-completions
		   (sccs-make-completions sc-file sc-dir
					  (list ad-return-value))))
      (let ((i-len (length item)))
	(when (>= i-len pf-len)
	  (setq item (substring item 0 pf-len)
		i-len pf-len)
	  (when (string-equal item sc-file)
	    (let ((x 0))
	      (while (and (< x (min rv-len i-len))
			  (eq (aref item x)
			      (aref ad-return-value x)))
		(setq x (1+ x)))
	      (when (< x rv-len)
		(setq ad-return-value (substring ad-return-value 0 x)
		      rv-len x)))))))))

(provide 'sccs-complete)


;;; Local Variables:
;;; mode: emacs-lisp
;;; prompt-to-byte-compile: nil
;;; end:
